{-
	Program for converting GHC ext-core representation to JSON

	Dependencies: json and extcore package.
-}

-- Core utilities
import Language.Core.Parser
import Language.Core.ParseGlue
import Language.Core.Core
import Language.Core.Encoding
import qualified Data.ByteString.Char8 as B
import Control.Monad.State.Strict
import Data.List

-- IO utilities
import System.Environment                   ( getArgs )

-- JSON utilities
import Text.JSON.Pretty
import Text.JSON.Generic


main = do 
    (inFile:outFile:_) <- getArgs
    compile inFile outFile


-- Parse inFile and write JSON output to outFile
compile :: FilePath -> FilePath -> IO ()
compile inFile outFile = do
    stuff <- liftIO $ B.readFile inFile
    let json = getJS $ parse (B.unpack stuff) 0;
		render = renderStyle $ Style PageMode 130 1
	in writeFile outFile $ render $ pp_value json


-- Generate JSON tree
getJS :: ParseResult Module -> JSValue
getJS (FailP err) = error ("error: " ++ err)
getJS (OkP m) = modJS m


-- Module
modJS :: Module -> JSValue
modJS (Module anmname tdef vdefg) =
    JSObject $ toJSObject
        [("%module", toJSON $ (zDecodeString . show $ anmname, "")),
         ("tdef", JSArray $ map tdefJS tdef),
         ("vdefg", JSArray $ map vdefgJS vdefg)]


-- Type definition
tdefJS :: Tdef -> JSValue
tdefJS (Data tcon tbinds cdefs) = 
    JSObject $ toJSObject
        [("%data", idJS tcon),
         ("tyvar", JSArray $ map tbindJS tbinds),
         ("cdef", JSArray $ map cdefJS cdefs)]
tdefJS (Newtype tcon1 tcon2 tbind ty) = 
    JSObject $ toJSObject
        [("%newtype", idJS tcon1), 
         ("qtycon", idJS tcon2), 
         ("tyvar", JSArray $ map tbindJS tbind),
         ("ty", tyJS ty)]


-- Constructor definition
cdefJS :: Cdef -> JSValue
cdefJS (Constr qual tbinds tys) = 
    JSObject $ toJSObject
        [("qdcon", idJS qual),
         ("tyvar", JSArray $ map tbindJS tbinds),
         ("aty", JSArray $ map tyJS tys)]


-- Value definition
vdefgJS :: Vdefg -> JSValue
vdefgJS (Rec vdefs) = 
    JSObject $ toJSObject 
        [("%rec", JSArray $ map vdefJS vdefs)]
vdefgJS (Nonrec vdef) = vdefJS vdef

vdefJS :: Vdef -> JSValue
vdefJS (Vdef (qvar, ty, exp)) =
    JSObject $ toJSObject
        [("qvar", idJS qvar),
         ("ty", tyJS ty),
         ("exp", expJS exp)]


-- Atomic expression && Expression && Argument
expJS :: Exp -> JSValue

expJS (Var qvar) = JSObject $ toJSObject 
    [("qvar", idJS qvar)]

expJS (Dcon qdcon) =  JSObject $ toJSObject 
    [("qdcon", idJS qdcon)]

expJS (Lit lit) = litJS lit

expJS (App exp1 exp2) = -- todo ???
    JSObject $ toJSObject
        [("aexp", expJS exp1),
         ("args", JSObject $ toJSObject $ 
            [("aexp", expJS exp2)] -- Application with value argument
        )]

expJS (Appt exp ty) = -- todo ???
    JSObject $ toJSObject
        [("aexp", expJS exp),
         ("args", JSObject $ toJSObject $ 
            [("aty", tyJS ty)]
        )] -- todo ? Application with type argument

expJS (Lam bind exp) =
    JSObject $ toJSObject
        [("lambda", binderJS bind),
         ("exp", expJS exp)] --todo ?

expJS (Let vdefg exp) =
    JSObject $ toJSObject
        [("%let", vdefgJS vdefg),
         ("%in", expJS exp)]

expJS (Case exp vbind ty alts) =
    JSObject $ toJSObject
        [("%case", tyJS ty),
         ("exp", expJS exp),
         ("%of", vbindJS vbind),
         ("alt", JSArray $ map altJS alts)]

expJS (Cast exp ty) =
    JSObject $ toJSObject
        [("%cast", expJS exp),
         ("aty", tyJS ty)]

expJS (Note string exp) =
    JSObject $ toJSObject
        [("%note", toJSON . zDecodeString $ string),
         ("exp", expJS exp)]

expJS (External string ty) =
    JSObject $ toJSObject
        [("%external ccal", toJSON . zDecodeString $ string), 
         ("aty", tyJS ty)]


-- Argument


-- Case alternative
altJS :: Alt -> JSValue
altJS (Acon qdcon tbinds vbinds exp) =
    JSObject $ toJSObject
        [("qdcon", idJS qdcon),
         ("tyvar", JSArray $ map tbindJS tbinds),
         ("vbind", JSArray $ map vbindJS vbinds),
         ("exp", expJS exp)]

altJS (Alit lit exp) =
    JSObject $ toJSObject
        [("lit", litJS lit),
         ("exp", expJS exp)]

altJS (Adefault exp) =
    JSObject $ toJSObject
        [("%_", expJS exp)]


-- Binder
binderJS :: Bind -> JSValue
binderJS (Vb vbind) =
    JSObject $ toJSObject
        [("vbind", vbindJS vbind)]
binderJS (Tb tbind) = 
    JSObject $ toJSObject
        [("tyvar", tbindJS tbind)]


-- Type binder
tbindJS :: Tbind -> JSValue
tbindJS (tvar, _) =
    toJSON . zDecodeString $ tvar


-- Value binder
vbindJS :: Vbind -> JSValue
vbindJS (var, ty) =
    JSObject $ toJSObject
        [("var", toJSON . zDecodeString $ var),
         ("ty", tyJS ty)]


-- Literal
litJS :: Lit -> JSValue
litJS (Literal corelit ty) = 
    JSObject $ toJSObject
        [("lit", corelitJS corelit),
         ("ty", tyJS ty)]

corelitJS :: CoreLit -> JSValue
corelitJS (Lint x)      =
    JSObject $ toJSObject
        [("int", toJSON $ x)]
corelitJS (Lrational x) =
    JSObject $ toJSObject 
        [("rational", JSRational True x)]
corelitJS (Lchar x)     =
    JSObject $ toJSObject 
        [("char", toJSON x)]
corelitJS (Lstring x)   =
    JSObject $ toJSObject 
        [("string", toJSON x)]


-- Atomic type && Basic type && Type
tyJS :: Ty -> JSValue
tyJS (Tvar tyvar) = 
    JSObject $ toJSObject 
        [("tyvar", toJSON . zDecodeString $ tyvar)]

tyJS (Tcon qtycon) = 
    JSObject $ toJSObject 
        [("qtycon", idJS qtycon)]

tyJS (Tapp ty1 ty2) =
    JSObject $ toJSObject
        [("bty", tyJS ty1),
         ("aty", tyJS ty2)]

tyJS (Tforall tbind ty) =
    JSObject $ toJSObject
        [("%forall", tbindJS tbind),
         ("ty", tyJS ty)]
tyJS (TransCoercion ty1 ty2) =
    JSObject $ toJSObject
        [("%trans", tyJS ty1),
         ("aty", tyJS ty2)]
tyJS (SymCoercion ty) =
    JSObject $ toJSObject
        [("%sym", tyJS ty)]
tyJS (UnsafeCoercion ty1 ty2) =
    JSObject $ toJSObject
        [("%unsafe", tyJS ty1),
         ("aty", tyJS ty2)]
tyJS (InstCoercion ty1 ty2) =
    JSObject $ toJSObject
        [("%inst", tyJS ty1),
         ("aty", tyJS ty2)]
tyJS (LeftCoercion ty) =
    JSObject $ toJSObject
        [("%left", tyJS ty)]
tyJS (RightCoercion ty) =
    JSObject $ toJSObject
        [("%right", tyJS ty)]


-- Identifier
idJS :: (Qual Tcon) -> JSValue
idJS (Just a, str) = toJSON $ (zDecodeString . show $ a, zDecodeString str)
idJS (Nothing, str) = toJSON . zDecodeString $ str
