#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Run PyHaskell benchmarks.
"""
from __future__ import print_function
from subprocess import check_call, check_output, Popen, PIPE
import os
import sys
import math
import argparse

from tools import zdecode, pypytrace


# Settings
GHC = 'ghc-7.4.2'
CUSTOM_GHC = 'ghc-7.7.20121016'
FORCE = '-fforce-recomp'
TARGET_VERSION = 5
BENCHMARK_FOLDER = 'src/'
CORE2JS = 'pyhaskell/core/core2js'


# Benchmark executable targets, compilation details
TARGET_INFO = {
    'ghc': (GHC, '-fasm -O0'),  # GHC NCG backend
    'ghc2': (GHC, '-fasm -O2'),  # GHC NCG backend, opt level 2
    'llvm': (GHC, '-fllvm -O0'),  # GHC LLVM backend
    'jit': ('bin/jit-c-v', '-fasm -O0 -fext-core'),  # PyHaskell with JIT
    'nojit': ('bin/nojit-c-v', '-fasm -O0 -fext-core'),  # PyHaskell no JIT
    'pyhaskell': ('python ../haskell-python/pyhaskell/main.py', ''),
}


# A list of available benchmarks
BENCHMARKS = ['addition', 'fibonacci', 'length', 'math']


def _compile_info(target, target_version=TARGET_VERSION):
    """Get compilation details for the target."""
    executable = target
    flags = ''

    # Precise target name
    if target in TARGET_INFO:
        executable, flags = TARGET_INFO[target]
        if target in ('jit', 'nojit'):
            executable += str(target_version)

    # jit/nojit target
    if target.startswith(TARGET_INFO['jit'][0]) or target == 'haskell-c':
        flags = TARGET_INFO['jit'][1:2]
    elif (target.startswith(TARGET_INFO['nojit'][0])
          or target == 'haskell-nojit-c'):
        flags = TARGET_INFO['nojit'][1:2]

    return executable, flags


def _parse_time_info(text):
    """Parse the output of /usr/bin/time command."""
    return [float(i.split(':', 1)[1].strip())
            for i in text.rsplit('\n', 4) if i]


def time_benchmark(binary, benchmark, priority=False):
    """Time a single run of a benchmark with a specfic target executable."""
    cmd = '/usr/bin/time -f '\
            '"\ntime: %e\ncontext switches: %c\nwaits: %w"'
    cmd += ' %s %s' % (binary, benchmark)

    if priority:
        cmd = 'chrt -f 99 ' + cmd

    #with Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True) as proc:
    proc = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
    text, warnings = proc.communicate()

    if proc.returncode:
        print("## Error, wrong returncode %i, aborting" % proc.returncode)
        print(cmd, text, warnings)
        sys.exit(1)

    return text.strip(), warnings.strip()


def compile_benchmark(executable, flags, benchmark, cmd=check_output):
    """Compile a benchmark with the given executable."""
    def _clean(path):
        os.remove(path + '.o')
        os.remove(path + '.hi')

    if benchmark in BENCHMARKS:
        benchmark = '%s%s%s' % (BENCHMARK_FOLDER, benchmark, '.hs')
    if not os.path.isfile(benchmark):
        print("## Error, unknown benchmark: %s" % benchmark)
        sys.exit(2)

    # GHC must always be compiled beforehand
    if executable.startswith('ghc') or executable == GHC:
        cmd('%s %s %s %s' % (executable, flags, FORCE, benchmark), shell=True)
        if os.path.splitext(benchmark)[1] == '.hs':
            benchmark = benchmark[:-3]  # Remove .hs
        _clean(benchmark)
        return '', benchmark

    # PyHaskell should be fed .hcj, for fairness sake
    path = os.path.splitext(benchmark)[0]
    hcj = '%s.hcj' % path
    if (not os.path.isfile(hcj) or
        os.path.getmtime(hcj) < os.path.getmtime(benchmark)):
        cmd('%s %s %s %s' % (CUSTOM_GHC, flags, FORCE, benchmark), shell=True)
        cmd('%s %s.hcr %s.hcj' % (CORE2JS, path, path), shell=True)
        os.remove(path)
        _clean(path)

    return executable, hcj


def run_benchmark(target, benchmark, runs, verbose=False, priority=False):
    """Run an individual benchmark to find mean average run time."""
    correct = None
    result = []
    for i in range(runs):
        output, time_info = time_benchmark(target, benchmark, priority)
        timings, switches, waits = _parse_time_info(time_info)
        result.append(timings)

        # Make sure we produce the same output in each run
        if correct is None:
            correct = output
        elif correct != output:
            print("## Error, two runs produced different output")
            print(i, correct, output)
            sys.exit(1)

        # Warn if context switches or waits are to high
        if verbose and (switches > 0 or waits > 1):
            print("# [%i] Warning, context switches: %i, waits: %i" %
                  (i, switches, waits))

    # Find the mean and sample standard deviation
    if runs < 1:
        mean, sample_std_dev, output = 0, 0, None
    elif runs == 1:
        mean, sample_std_dev = result[0], 0
    else:
        mean = sum(result) / len(result)
        dev = [i - mean for i in result]
        dev2 = [i ** 2 for i in dev]
        sample_std_dev = sum(dev2) / (len(result) - 1)
    return mean, math.sqrt(sample_std_dev), correct


def zdecode_benchmark(path):
    """Write out a z-decoded benchmark."""
    path = os.path.splitext(path)[0]
    with open(path + '.hcr', 'r') as f:
        text = f.read()
    with open(path + '.z.hcr', 'w') as f:
        f.write(zdecode.z_decode(text))


def count_jitlog(output, *args):
    """Count operations in jit trace-log."""
    os.environ['PYPYLOG'] = 'jit-log-opt:jit.log'
    results = run_benchmark(*args)
    if output is not None:
        assert output == results[2]

    with open('jit.log', 'r') as f:
        ops, guards = pypytrace.count(f.read())
    print("## Trace operations: %i  Guards: %i" % (ops, guards))


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('target', nargs='?', default='jit',
                        help='target binary to benchmark')
    parser.add_argument('benchmark', nargs='*', default='all',
                        help='a specific benchmark to run')
    parser.add_argument('-t', '--tv', nargs='?', default=TARGET_VERSION,
                        type=int, help='PyHaskell target version')
    parser.add_argument('-n', '--num', nargs='?', default=3, type=int,
                        help='number of times to repeat benchmark')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='print detailed information')
    parser.add_argument('-p', '--priority', action='store_true', default=False,
                        help='run benchmark at highest priority, require root')
    parser.add_argument('-z', '--zdecode', action='store_true', default=False,
                        help='create z-decoded hcr file of benchmark')
    parser.add_argument('-c', '--count', action='store_true', default=False,
                        help='count operations in jit trace-log')

    ns = parser.parse_args()
    benchmarks = BENCHMARKS if ns.benchmark == 'all' else ns.benchmark
    executable, flags = _compile_info(ns.target, ns.tv)

    cmd = check_call if ns.verbose else check_output
    cmd('make', shell=True)  # Compile core2js if needed

    for name in benchmarks:
        # Disable pypylog for benchmarks
        old_value = os.environ['PYPYLOG'] if 'PYPYLOG' in os.environ else None
        os.environ['PYPYLOG'] = ''

        target, path = compile_benchmark(executable, flags, name, cmd)
        mean, stddev, output = run_benchmark(target, path,
                                     ns.num, ns.verbose, ns.priority)

        print()
        print("## %s output: %s" % (path, output))
        print(u"## Mean average of %i runs: %.4f seconds (\u00B1 %.4f)" % (
            ns.num, mean, 1.96 * stddev))

        if ns.count and ns.target not in ('ghc', 'ghc2', 'llvm', 'nojit'):
            count_jitlog(output, target, path, 1)
        if ns.zdecode:
            zdecode_benchmark(path)

        if old_value is not None:
            os.environ['PYPYLOG'] = old_value


if __name__ == '__main__':
    main()
