# -*- coding: utf-8 -*-
from pypy.rlib import jit, unroll


class Symbol(object):
    """ A cached symbol that can be compared by identity (which is not true for
    strings). """
    all_symbols = {}

    def __init__(self, str_):
        self.str = str_

    @staticmethod
    def get_symbol(str):
        if str in Symbol.all_symbols:
            return Symbol.all_symbols[str]
        result = Symbol(str)
        Symbol.all_symbols[str] = result
        return result

    def tostr(self):
        return self.str

    def __str__(self):
        return self.tostr()


class HaskellObject(object):
    """ Base class for all objects that the interpreter handles. """
    _attrs_ = []

    def step(self, todo):
        """ Perform and interpretation step where self should move in the
        direction of being more evaluated. """
        raise NotImplementedError

    def getvalue(self):
        """ Dereference self to be a value. Can return None if self is not
        evaluated."""
        return None

    def substitute(self, subst):
        """ Perform a substitution. """
        raise NotImplementedError

    def enumerate_head(self, subst):
        """ Enumerate the variables in self. """
        raise NotImplementedError

    def replace_vars(self, subst, occurrences):
        """ Replace varibles in self with what is found in subst."""
        raise NotImplementedError

    def __str__(self):
        return self.tostr()


class Value(HaskellObject):
    """ Base class for evaluated values (i.e. already in head-normal form). """
    _attrs_ = []

    def step(self, todo):
        """ self is already a value. Continue with the todo stack."""
        return todo.step(self)

    def getvalue(self):
        """ self is a value, just return it. """
        return self

    def substitute(self, subst):
        return self

    def enumerate_head(self, subst):
        return self

    def replace_vars(self, subst, occurrences):
        return self

    def tostr(self):
        raise NotImplementedError


NO_MATCH = 0
DEFINITE_MATCH = 1
NEEDS_HNF = 2


class Constructor(Value):
    """ A constructor. This is an abstract base class, there are
    subclasses generated below for various numbers of arguments."""
    _immutable_fields_ = ["symbol"]

    def __init__(self, symbol):
        assert isinstance(symbol, Symbol)
        self.symbol = symbol

    @jit.unroll_safe
    def match(self, other, subst):
        value = other.getvalue()
        if value is None:
            return NEEDS_HNF
        assert isinstance(value, Constructor)
        if self.symbol is not value.symbol:
            return NO_MATCH
        for i in range(self.numargs()):
            arg1 = self.getarg(i)
            assert isinstance(arg1, NumberedVar)
            arg2 = value.getarg(i)
            res = arg1.match(arg2, subst)
            assert res == DEFINITE_MATCH
        return DEFINITE_MATCH

    @jit.unroll_safe
    def substitute(self, subst):
        args = [None] * self.numargs()
        for i in range(self.numargs()):
            args[i] = self.getarg(i).substitute(subst)
        return make_constructor(self.symbol, args)

    def enumerate_head(self, subst):
        args = [None] * self.numargs()
        for i in range(self.numargs()):
            args[i] = self.getarg(i).enumerate_head(subst)
        return make_constructor(self.symbol, args)

    def replace_vars(self, subst, occurrences):
        args = [None] * self.numargs()
        for i in range(self.numargs()):
            args[i] = self.getarg(i).replace_vars(subst, occurrences)
        return make_constructor(self.symbol, args)

    def __eq__(self, other):
        return (isinstance(other, Constructor) and
                self.symbol is other.symbol and
                self.getargs() == other.getargs())

    def __ne__(self, other):
        return not (self == other)

    def tostr(self):
        args = ", ".join([arg.tostr() for arg in self.getargs()])
        if args:
            args = " " + args
        return "(Constr %s%s)" % (self.symbol.tostr(), args)

    @jit.unroll_safe
    def getargs(self):
        args = [None] * self.numargs()
        for i in range(self.numargs()):
            args[i] = self.getarg(i)
        return args

    def getname(self):
        return self.symbol.tostr()


def make_arg_subclass(n, base):
    rangen = unroll.unrolling_iterable(range(n))

    class cls(base):
        _immutable_fields_ = ["arg%s" % i for i in rangen]

        def __init__(self, function, args):
            base.__init__(self, function)
            for i in rangen:
                setattr(self, "arg%s" % i, args[i])
            assert len(args) == n

        def numargs(self):
            return n

        def getarg(self, i):
            assert i < n
            for j in rangen:
                if i == j:
                    return getattr(self, "arg%s" % j)
            assert 0

    cls.__name__ = "%s%s" % (base.__name__, n)
    return cls


specialized_constr_classes = [
    make_arg_subclass(i, Constructor) for i in range(11)]


class ConstructorN(Constructor):
    _immutable_fields_ = ["args[*]"]

    def __init__(self, function, args):
        Constructor.__init__(self, function)
        self.args = args[:]

    def numargs(self):
        return len(self.args)

    def getarg(self, i):
        return self.args[i]


def make_constructor(function, args):
    if len(args) < len(specialized_app_classes):
        cls = specialized_constr_classes[len(args)]
    else:
        cls = ConstructorN
    return cls(function, args)


class AbstractFunction(Value):
    _immutable_fields_ = ["name"]

    def __init__(self, name):
        self.name = name

    def apply(self, application, todo):
        raise NotImplementedError


class Function(AbstractFunction):
    """ A user-defined function, i.e. written in Haskell. """
    _immutable_fields_ = ["rules[*]"]

    def __init__(self, name, rules):
        AbstractFunction.__init__(self, name)
        self.rules = rules
        if rules is not None:
            jit.promote(self.rules)

    def tostr(self):
        if self.rules is None:
            return "ForwardReference(" + self.name + ")"
        else:
            return '(%s %s)' % (
                    self.name, " ".join([a.tostr() for a in self.rules]))

    def become(self, function):
        self.rules = function.rules
        jit.promote(self.rules)

    @jit.unroll_safe
    def apply(self, application, todo):
        assert self.rules is not None  # ForwardReference probably
        for i in range(len(self.rules)):
            rule = self.rules[i]
            subst, todo = rule.apply(application, todo)
            if subst is not None:
                return subst, todo
        assert 0


class Rule(object):
    """ One rule of a user-defined function. """
    _immutable_fields_ = ["patterns[*]", "body", "numvars",
                          "occurrences[*]", "recursive"]

    def __init__(self, patterns, body, recursive):
        subst = {}
        self.patterns = [pattern.enumerate_head(subst) for pattern in patterns]
        occurrences = [0] * len(subst)
        self.body = body.replace_vars(subst, occurrences)
        self.recursive = recursive
        self.numvars = len(subst)
        self.occurrences = occurrences

    @jit.unroll_safe
    def apply(self, application, todo):
        subst = [None] * self.numvars
        for i in range(application.numargs()):
            arg = application.getarg(i)
            match = self.patterns[i].match(arg, subst)
            if match == NO_MATCH:
                return None, todo
            elif match == NEEDS_HNF:
                return application.eval_after(arg, i, todo)
            else:
                assert match == DEFINITE_MATCH

        # introduce thunks necessary for sharing
        for i in range(len(subst)):
            val = subst[i].getvalue()
            if val is not None:
                subst[i] = val
            elif (isinstance(subst[i], Application) and
                  self.occurrences[i] > 1):
                subst[i] = Thunk(subst[i])
        res = Substitution(self.body, subst, self.recursive)

        return res, todo

    def tostr(self):
        return "(%s)" % " ".join([arg.tostr() for arg in self.patterns])

    def __str__(self):
        return self.tostr()


class Substitution(HaskellObject):
    """ The body of a function with numbered variables substituted by values.
    """
    _immutable_fields_ = ["rhs", "subst", "recursive"]

    def __init__(self, rhs, subst, recursive):
        self.rhs = rhs
        self.subst = subst
        self.recursive = recursive

    def apply(self):
        jit.promote(self.rhs)
        return self.rhs.substitute(self.subst)


class PrimFunction(AbstractFunction):
    """ A primitive function, i.e. one not implemented in Haskell but at the
    machine level. """
    _immutable_fields_ = ["prim_function", "arity", "strictargs"]

    def __init__(self, name, function, arity, strictargs):
        AbstractFunction.__init__(self, name)
        self.prim_function = function
        self.arity = arity
        self.strictargs = strictargs

    @jit.unroll_safe
    def apply(self, application, todo):
        args = [None] * self.arity
        if application.numargs() < self.arity:
            return self, todo

        assert application.numargs() == self.arity
        for i in range(self.arity):
            arg = application.getarg(i)
            strictarg = self.strictargs[i]
            if strictarg:
                value = arg.getvalue()
                if value is None:
                    return application.eval_after(arg, i, todo)
                args[i] = value
            else:
                args[i] = arg
        return self.prim_function(args), todo

    def tostr(self):
        return self.name


class Var(HaskellObject):
    _immutable_fields_ = ["name"]

    def __init__(self, name):
        self.name = name

    def match(self, obj, subst):
        raise NotImplementedError

    def substitute(self, subst):
        raise NotImplementedError

    def enumerate_head(self, subst):
        res = NumberedVar(len(subst), self.name)
        subst[self] = res
        return res

    def replace_vars(self, subst, occurrences):
        res = subst[self]
        occurrences[res.index] += 1
        return res

    def tostr(self):
        return "Var: " + self.name


class NumberedVar(HaskellObject):
    _immutable_fields_ = ["name", "index"]

    def __init__(self, index, name=""):
        self.index = index
        self.name = name

    def match(self, obj, subst):
        assert subst[self.index] is None
        subst[self.index] = obj
        return DEFINITE_MATCH

    def substitute(self, subst):
        arg = subst[self.index]
        value = arg.getvalue()
        if value is not None:
            arg = value
        return arg

    def tostr(self):
        return self.name


class Application(HaskellObject):
    """ A function application. This is an abstract base class, there are
    subclasses generated below for various numbers of arguments."""
    _immutable_fields_ = ["function"]

    def __init__(self, function):
        self.function = function

    def step(self, todo):
        value = self.function.getvalue()
        if value:
            jit.promote(value)
            assert isinstance(value, AbstractFunction)
            return value.apply(self, todo)

        # higher order
        return self.eval_after(self.function, -1, todo)

    def eval_after(self, arg, i, todo):
        todo = CopyStackElement(self, i, todo)
        return arg, todo

    def replace(self, index, value):
        if index == -1:
            return make_application(value, self.getargs())
        args = self.getargs()
        jit.promote(index)
        args[index] = value
        return make_application(self.function, args)

    @jit.unroll_safe
    def substitute(self, subst):
        args = [None] * self.numargs()
        for i in range(self.numargs()):
            args[i] = self.getarg(i).substitute(subst)
        res = make_application(self.function.substitute(subst), args)
        return res

    def replace_vars(self, subst, occurrences):
        args = [self.getarg(i).replace_vars(subst, occurrences)
                    for i in range(self.numargs())]
        return make_application(self.function.replace_vars(subst, occurrences),
                                args)

    def tostr(self):
        tmp = self.function.tostr()
        if self.numargs() > 0:
            tmp += " " + " ".join(
                [self.getarg(i).tostr() for i in range(self.numargs())])
        return tmp

    @jit.unroll_safe
    def getargs(self):
        args = [None] * self.numargs()
        for i in range(self.numargs()):
            args[i] = self.getarg(i)
        return args

    def enumerate_head(self, subst):
        """An application has no variables to enumerate, return self"""
        return self


specialized_app_classes = [
    make_arg_subclass(i, Application) for i in range(11)]


class ApplicationN(Application):
    _immutable_fields_ = ["args[*]"]

    def __init__(self, function, args):
        Application.__init__(self, function)
        self.args = args

    def numargs(self):
        return len(self.args)

    def getarg(self, i):
        return self.args[i]


def make_application(function, args):
    if len(args) < len(specialized_app_classes):
        cls = specialized_app_classes[len(args)]
    else:
        cls = ApplicationN
    return cls(function, args)


class Thunk(HaskellObject):
    """ An unevaluated function application. """
    def __init__(self, application):
        assert isinstance(application, Application)
        self.application = application

    def step(self, todo):
        app = self.application
        if isinstance(app, Value):
            return app, todo
        return self.application, UpdateStackElement(self, todo)

    def getvalue(self):
        app = self.application
        if isinstance(app, Value):
            return app
        return None

    def substitute(self, subst):
        raise NotImplementedError

    def tostr(self):
        if isinstance(self.application, Value):
            return self.application.tostr()
        return "?%s" % self.application


class PartialApp(HaskellObject):
    _immutable_fields_ = ["function", "args[*]", "arity"]

    def __init__(self, app, args):
        self.function = app
        self.args = args

        if isinstance(app, Function):
            self.arity = len(app.rules[0].patterns)
        elif isinstance(app, PrimFunction):
            self.arity = app.arity
        else:
            raise NotImplementedError

    def getargs(self):
        return self.args

    def tostr(self):
        return "(partialapp: %s)" % self.function

    def __str__(self):
        return self.tostr()


# _________________________________________________________________
# head normal form stackless

def get_printable_location(function):
    if function is None:
        return "None"
    return function.tostr()

jitdriver = jit.JitDriver(
        greens=["function"],
        reds=["todo", "expr"],
        get_printable_location=get_printable_location,
        )


class StackElement(object):
    """ Base class of the stack elements of the evaluation stack. The content
    of the stack roughly corresponds to the Launchbury semantics. """

    def __init__(self, next):
        self.next = next

    def step(self, value):
        raise NotImplementedError("abstract base class")


class CopyStackElement(StackElement):
    """ Need to copy the top of the stack. """

    def __init__(self, application, index, next):
        StackElement.__init__(self, next)
        self.application = application
        self.index = index

    def step(self, value):
        return self.application.replace(self.index, value), self.next

    def __str__(self):
        return "copy(%s, %s, %s)" % (self.application, self.index, self.next)


class UpdateStackElement(StackElement):
    """ Need to update the thunk stored in this after its content has been
    evaluated. """

    def __init__(self, thunk, next):
        StackElement.__init__(self, next)
        self.thunk = thunk

    def step(self, value):
        assert self.thunk.getvalue() is None
        self.thunk.application = value
        return value, self.next

    def __str__(self):
        return "upd(%s, %s)" % (self.thunk, self.next)


def evaluate_hnf(obj):
    assert isinstance(obj, Application)
    return main_loop(obj)


def main_loop(expr):
    function = None
    todo = None
    while True:
        jitdriver.jit_merge_point(function=function, todo=todo, expr=expr)
        if isinstance(expr, Substitution):
            expr = expr.apply()
        if isinstance(expr, Value) and todo is None:
            break
        #print expr, todo
        #import pdb; pdb.set_trace()
        #print ""
        #print expr.tostr(), "    => "
        expr, todo = expr.step(todo)
        function = None
        if isinstance(expr, Substitution):
            recursive = expr.recursive
            #print recursive
            function = expr.rhs
            #print repr(function)
            #print function.name
            if recursive:
                #print "can enter jit", function, expr
                jitdriver.can_enter_jit(function=function,
                                        todo=todo, expr=expr)
    return expr
