# -*- coding: utf-8 -*-
from pypy.rlib.parsing.ebnfparse import parse_ebnf, make_parse_function
from pypy.rlib.parsing.tree import RPythonVisitor, Nonterminal
from pypy.rlib.streamio import open_file_as_stream

from pyhaskell.interpreter.module import CoreMod
from pyhaskell.interpreter.primtype import Int, Double, Char, Addr
from pyhaskell.interpreter.haskell import (
        Function, PrimFunction, Rule, PartialApp, Symbol,
        HaskellObject, Var, AbstractFunction, Value,
        make_application, make_constructor)


# EBNF for JSON parser
EBNF = """
    STRING: "\\"([^\\"\\\\]|(\\\\.))*\\"";
    NUMBER: "\-?(0|[1-9][0-9]*)(\.[0-9]+)?([eE][\+\-]?[0-9]+)?";
    IGNORE: "[ \t\n]";
    value: <STRING> | <NUMBER> | <object> | <array> |
           <"null"> | <"true"> | <"false">;
    object: ["{"] (pair [","])* pair ["}"] | ["{}"];
    array: ["["] (value [","])* value ["]"] | ["[]"];
    pair: STRING [":"] value;
    """

REGEXS, RULES, TO_AST = parse_ebnf(EBNF)
PARSE = make_parse_function(REGEXS, RULES, eof=True)


class AST(RPythonVisitor):
    """Class for traversing and generating AST.

    This class traverses the JSON datastructure and generates an AST.
    """

    def __init__(self):
        self.mident = ""

        # Recursive needs to be a stack due to let expressions
        self.recursive = []

        # These stacks are necessary to keep
        # variables in scope in case-expressions
        self.exp_stack = []

    def get_ast(self, tree):
        """Go through the tree and populate the module it represents."""
        self.visit_module(tree)
        return CoreMod.all_modules[self.mident]

    def _module(self, node):
        """Returns a module name and reference."""
        if not isinstance(node, Nonterminal):
            mident = self.mident
            name = get_string(node)
        else:
            mident = get_string(node.children[0])
            name = get_string(node.children[1])
        return mident, name

    def visit_object(self, node):
        """Visit a generic object."""
        raise NotImplementedError("Node type not supported: %s" %
                                  get_string(node.children[0].children[0]))

    def visit_module(self, node):
        """Create a new module."""
        self.mident = get_string(node.children[0].children[1].children[0])
        CoreMod(self.mident)

        # Create type definitions
        for child in node.children[1].children[1].children:
            child_type = get_string(child.children[0].children[0])
            if child_type == '%data':
                self.visit_constructors(child)
            else:
                raise NotImplementedError('%s unsupported' % child_type)

        # Create value definitions
        for child in node.children[2].children[1].children:
            self.visit_recursive_value_defs(child)

    def visit_constructors(self, node):
        """Visit a type constructor with data constructors."""
        # Create zero or more type variables
        #type_vars = [get_string(n) for n in
        #                node.children[1].children[1].children]
        for child in node.children[1].children[1].children:
            get_string(child)

        # Create data constructors
        for child in node.children[2].children[1].children:
            mident, name = self._module(child.children[0].children[1])
            arity = len(child.children[2].children[1].children)
            get_module(mident).add_data_constr(name, arity)

    def visit_recursive_value_defs(self, node):
        """Visit either recursive definitions or one single definition."""
        type_ = get_string(node.children[0].children[0])
        if type_ == '%rec':
            self.recursive.append(True)
            for child in node.children[0].children[1].children:
                self.visit_value_definition(child)
            self.recursive.pop()
        else:
            self.visit_value_definition(node)

    def visit_expression(self, node):
        """Visit an expression node (exp) or atomic expr (aexp)."""
        type_ = get_string(node.children[0].children[0])
        #print "EXP:", type_, len(node.children)

        # Atomic expressions
        if type_ == 'qvar':
            expr = get_variable(*self._module(node.children[0].children[1]))
        elif type_ == 'qdcon':
            mod, name = self._module(node.children[0].children[1])
            expr = get_module(mod).create_data_expr(name)
        elif type_ == 'qtycon':
            print node  # XXX
            raise NotImplementedError
        elif type_ == 'lit':
            expr = self.visit_literal(node.children[0].children[1])
        elif type_ == 'aexp' and len(node.children) == 1:
            expr = self.visit_expression(node.children[0].children[1])

        # Application
        elif type_ == 'aexp' and len(node.children) == 2:
            expr = self.visit_expression(node.children[0].children[1])
            child = node.children[1].children[1].children[0].children[0]
            if get_string(child) != 'aty':
                arg = self.visit_expression(node.children[1].children[1])
                # Hack to remove data aliases when it is a literal
                if (isinstance(expr, Function) and isinstance(arg, Value)
                    and expr.name in ('I#', 'C#')):
                    expr = arg
                else:
                    expr = make_partial_app(expr, [arg])

        # Lambda expression
        elif type_ == 'lambda':
            child = node.children[0].children[1].children[0].children[1]
            var = self.visit_variable(child)
            self.exp_stack.append(var)
            expr = self.visit_expression(node.children[1].children[1])
            if not isinstance(expr, AbstractFunction):
                expr = Function("lambda", [Rule(self.exp_stack, expr, False)])
            self.exp_stack.pop()

        # Case expression
        elif type_ == '%case':
            # Note: A case expression is reduced to the
            # application of an expression to a function
            exp = self.visit_expression(node.children[1].children[1])

            # bind "of" to "exp"
            child = node.children[2].children[1].children[0].children[1]
            mod, name = self._module(child)
            exp = add_variable(mod, name, exp)

            # Create case alternatives
            alternatives = []
            for child in node.children[3].children[1].children:
                alternatives.append(self.handle_case_alternative(child))
            rules = []
            alternatives.reverse()
            for pat, expr in alternatives:
                pattern = [pat] + self.exp_stack
                rec = bool(self.recursive)
                rules.append(Rule(pattern, expr, rec))

            # Since "exp" is the scrutinee, we need it at the beginning
            # of the arguments to be applied.
            args = [exp] + self.exp_stack
            expr = make_partial_app(Function("Case", rules), args)

        else:
            # let, cast, note, label etc not implemented yet
            raise NotImplementedError("Expr type not supported: %s" % type_)

        assert isinstance(expr, HaskellObject)
        return expr

    def handle_case_alternative(self, node):
        """Visit a case alternative node.

        Method is named handle_ as it has a different signature than
        methods starting with visit_.
        """
        type_ = get_string(node.children[0].children[0])

        # Default case-alternative
        if type_ == '%_':
            pattern = Var('_')
            expr = self.visit_expression(node.children[0].children[1])

        # Literal alternative
        elif type_ == 'lit':
            child = node.children[0].children[1].children[0].children[1]
            pattern = self.visit_literal(child)
            expr = self.visit_expression(node.children[1].children[1])

        # Constructor alternative
        else:
            # Constructor value binders
            vbinds = []
            for child in node.children[2].children[1].children:
                vbinds.append(self.visit_variable(child))

            self.exp_stack.extend(vbinds)
            expr = self.visit_expression(node.children[3].children[1])
            start = len(self.exp_stack) - len(vbinds)
            assert start >= 0  # RPython require proving start is non-negative
            del self.exp_stack[start:]

            mod, name = self._module(node.children[0].children[1])
            module = get_module(mod)

            # TODO: temp hack to allow case alternative
            # where qdcon is a primitive type alias
            if name in module.data_aliases:
                pattern, = vbinds
            else:
                data_constr = module.qdcons[name]
                assert isinstance(data_constr, AbstractFunction)
                symbol = Symbol.get_symbol(data_constr.name)
                pattern = make_constructor(symbol, vbinds)

        assert isinstance(pattern, HaskellObject)
        assert isinstance(expr, HaskellObject)
        return pattern, expr

    def visit_literal(self, node):
        """Returns an instance of a literal."""
        type_ = get_string(node.children[0].children[0])
        value = get_string(node.children[0].children[1])

        if type_ == 'int':
            return Int(int(value))
        elif type_ == "rational":
            return Double(float(value))
        elif type_ == "char":
            return Char(value)
        else:
            return Addr(value)

    def visit_variable(self, node):
        """Visit a variable definition node."""
        mod, name = self._module(node.children[0].children[1])
        var = Var(name)
        return add_variable(mod, name, var)

    def visit_value_definition(self, node):
        """Visit a value definition which returns an expression."""
        expr = self.visit_expression(node.children[2].children[1])
        mod, name = self._module(node.children[0].children[1])
        return add_variable(mod, name, expr)


def make_partial_app(app, args):
    if isinstance(app, PartialApp):
        arity = app.arity
        args = app.getargs() + args
        app = app.function
    elif isinstance(app, PrimFunction):
        arity = app.arity
    elif isinstance(app, Function) and app.rules is not None:
        arity = len(app.rules[0].patterns)
    else:
        arity = 1

    if len(args) < arity:
        return PartialApp(app, args)
    else:
        return make_application(app, args)


def get_string(node):
    """Get the string value from a string node."""
    return ''.join(node.additional_info.split('"'))


def get_variable(mident, name):
    module = get_module(mident)
    if not name in module.qvars:
        module.qvars[name] = Function(name, None)
    return module.qvars[name]


def add_variable(mident, name, var):
    mod = get_module(mident)

    # Var might already exists as a "ForwardReference" function
    # if the function is recursive. Then we update obj with var's
    # rules and throw away var
    obj = mod.qvars.get(name, None)
    if isinstance(obj, Function) and isinstance(var, Function):
        if obj.rules is None:
            obj.become(var)
            return obj

    mod.qvars[name] = var
    return var


def get_module(name):
    """Returns module 'name', load from external JSCore file if necessary."""
    if name in CoreMod.all_modules:
        result = CoreMod.all_modules[name]
    else:
        print "Loading GHC module from hcj:", name
        assert False
        module_path = name.replace('.', '/').replace(':', '/')
        js = parse_js('./ghc_modules/%s.hcj' % module_path)
        result = AST().get_ast(js)
    assert isinstance(result, CoreMod)
    return result


def parse_js(path):
    """Parse JSCore file."""
    f = open_file_as_stream(path)
    doc = f.readall()
    f.close()
    t = PARSE(doc)
    t = TO_AST().transform(t)
    return t
