# -*- coding: utf-8 -*-
from pypy.rlib.jit import elidable

from pyhaskell.interpreter.haskell import (
    Value, PrimFunction, Symbol, Var, Function, Rule, make_constructor)


class CoreMod(object):
    """Represent a Haskell module."""
    _immutable_fields_ = ["mident"]

    all_modules = {}

    def __init__(self, mident):
        CoreMod.all_modules[mident] = self
        self.mident = mident
        self.qvars = {}  # Variables and functions
        self.qtycons = {}  # Type constructors
        self.qdcons = {}  # Data constructors
        self.data_aliases = {}  # Aliases for primitive types

    def __str__(self):
        return self.mident

    def add_data_constr(self, name, arity):
        """Create and register a data constructor.

        'name' is the name for the constructor,
        'arity' is the number of arguments the constructor expects.
        """
        vars_ = [Var('v%s' % str(i)) for i in range(arity)]
        constr = make_constructor(Symbol.get_symbol(name), vars_)
        con = Function(name, [Rule(vars_, constr, False)])
        self.qdcons[name] = con

    def add_primitive_alias(self, name, primitive):
        """Add a data constructor that is an alias for a primitive type."""
        assert issubclass(primitive, Value)
        self.data_aliases[name] = primitive

    def create_data_expr(self, name):
        if name in self.data_aliases:
            var = Var('_')
            return Function(name, [Rule([var], var, False)])
        else:
            return self.qdcons[name]


@elidable
def get_var(module, name):
    return module.qvars[name]


def expose_var(module, name, arg_types):
    return _expose(module.qvars, name, arg_types)


def expose_data_constr(module, name, arg_types):
    return _expose(module.qdcons, name, arg_types)


def expose_type_constr(module, name, arg_types):
    return _expose(module.qtycons, name, arg_types)


def _expose(storage, name, arg_types):
    from pypy.rlib import unroll
    """NOT_RPYTHON"""
    def decorator(function):
        """NOT_RPYTHON"""
        # some rpython magic...
        arity = len(arg_types)
        arityrange = unroll.unrolling_iterable(range(arity))

        def wrapped_f(args):
            # Check that arguments are instances of valid classes
            assert len(args) == arity
            tupleargs = ()
            for i in arityrange:
                x = args[i]
                assert isinstance(x, arg_types[i])
                tupleargs += (x, )
            return function(*tupleargs)

        func = PrimFunction(name, wrapped_f, arity, [True] * arity)
        storage[name] = func
        return func
    return decorator
