# -*- coding: utf-8 -*-
from pyhaskell.interpreter.primtype import Int, Float, Double
from pyhaskell.interpreter.haskell import (
    HaskellObject, Rule, Function, Symbol, Thunk, PrimFunction,
    Var, NumberedVar, NEEDS_HNF, evaluate_hnf, make_application,
    make_constructor, PartialApp)


class ForwardReference(HaskellObject):
    def __init__(self, name=""):
        self.name = name

    def become(self, x):
        self.__class__ = x.__class__
        self.__dict__.update(x.__dict__)

    def replace_vars(self, subst, _):
        return self

    def tostr(self):
        return "ForwardReference(" + self.name + ")"

    def __str__(self):
        return self.tostr()


def function(name, rules, recursive=False):
    rules = [Rule(pattern, body, recursive) for pattern, body in rules]
    return Function(name, rules)


def constr(name, *args):
    assert all([isinstance(arg, HaskellObject) for arg in args])
    return make_constructor(Symbol.get_symbol(name), list(args))


def app(func, *args):
    return make_application(func, list(args))


def thunk(func, *args):
    return Thunk(make_application(func, list(args)))


def cons(a, b):
    return constr("Cons", a, b)


w_true = constr("True")
w_false = constr("False")
w_nil = constr("Nil")


def expose_primitive(arity, strictargs=None):
    if strictargs is None:
        strictargs = [True] * arity

    def f(primitive):
        return PrimFunction(f.func_name, primitive, arity, strictargs)

    return f


@expose_primitive(2)
def prim_add(args):
    a0, a1 = args
    assert isinstance(a0, Int)
    assert isinstance(a1, Int)
    return Int(a0.value + a1.value)


@expose_primitive(2)
def prim_sub(args):
    a0, a1 = args
    assert isinstance(a0, Int)
    assert isinstance(a1, Int)
    return Int(a0.value - a1.value)


@expose_primitive(2)
def prim_gt(args):
    a0, a1 = args
    assert isinstance(a0, Int)
    assert isinstance(a1, Int)
    if a0.value > a1.value:
        return w_true
    return w_false


@expose_primitive(2)
def prim_eq(args):
    a0, a1 = args
    assert isinstance(a0, Int)
    assert isinstance(a1, Int)
    if a0.value == a1.value:
        return w_true
    return w_false


@expose_primitive(2, strictargs=[True, False])
def prim_seq(args):
    #assert args[0].getvalue() is not None
    return args[1]


x = Var("x")
id_ = function("id", [([x], x)])  # id x = x
dup = function("dup", [([x], constr("Pair", x, x))])
not_ = function("not", [([w_true], w_false), ([w_false], w_true)])
and_ = function("and", [([w_false, x], w_false), ([w_true, x], x)])
succ = function("succ", [([x], app(prim_add, x, Int(1)))])


hd = Var("hd")
tl = Var("tl")

append = ForwardReference()
append.become(function("append", [
                ([w_nil, x], x),
                ([cons(hd, tl), x], cons(hd, app(append, tl, x)))]))

left = Var("left")
right = Var("right")
value = Var("value")

l1 = Var("l1")
r1 = Var("r1")
v1 = Var("v1")
l2 = Var("l2")
r2 = Var("r2")
v2 = Var("v2")

f = Var("f")

w_empty = constr("Empty")


def node(left, value, right):
    return constr("Node", left, value, right)


map_ = ForwardReference()
map_.become(function("map", [
    ([f, w_nil], w_nil),
    ([f, cons(hd, tl)], cons(app(f, hd), app(map_, f, tl)))]))


treemap = ForwardReference()
treemap.become(function("treemap", [
    ([f, w_empty], w_empty),
    ([f, node(left, value, right)],
         node(app(treemap, f, left),
              app(f, value),
              app(treemap, f, right)))]))


eq_int_tree = ForwardReference()
eq_int_tree.become(function("eq_int_tree", [
    ([w_empty, w_empty], w_true),
    ([node(l1, v1, r1), node(l2, v2, r2)],
     app(and_, app(prim_eq, v1, v2),
         app(and_, app(eq_int_tree, l1, l2), app(eq_int_tree, r1, r2)))),
    ([l1, r1], w_false)]))


length_ = ForwardReference()
length_.become(function("length", [
    ([w_nil], Int(0)),
    ([cons(hd, tl)], app(prim_add, Int(1), app(length_, tl)))]))


acc = Var("acc")

length_iter = ForwardReference()
length_iter.become(function("length-iter", [
    ([w_nil, acc], acc),
    ([cons(hd, tl), acc],
     app(length_iter,
                      tl, app(prim_add,
                                              Int(1), acc)))]))


# corresponding haskell function

# length-seq-iter []     acc = acc
# length-seq-iter (x:xs) acc =
#   let acc' = 1+acc
#   in
#   seq acc' (length-seq-iter xs acc')

# lambda lifted version

# length-seq-iter []     acc = acc
# length-seq-iter (x:xs) acc = length-seq-iter-lambda xs (1+acc)
#
# length-seq-iter-lambda xs acc =
#   seq acc (length-seq-iter xs acc)

xs = Var("xs")
ys = Var("ys")

length_seq_iter = ForwardReference()
length_seq_iter_lambda = ForwardReference()
length_seq_iter.become(function("length-seq-iter", [
    ([w_nil, acc], acc),
    ([cons(hd, tl), acc],
     app(length_seq_iter_lambda,
                      tl, app(prim_add,
                                              Int(1), acc)))], recursive=True))
length_seq_iter_lambda.become(function("length-seq-iter-lambda", [
    ([xs, acc], app(prim_seq,
                                 acc, app(length_seq_iter, xs, acc)))]))


# let ones = 1 : ones in head ones  # XXX how do we express this?

# reverse [] = []
# reverse (x : xs) = reverse xs ++ [x]

reverse = ForwardReference()
reverse.become(function("reverse",
    [([w_nil], w_nil),
     ([cons(hd, tl)],
        app(append, app(reverse, tl),
            cons(hd, w_nil)))]))

# ifThenElse True  x _ = x
# ifThenElse False _ x = x

x = Var("x")
y = Var("y")
ifthenelse = function("ifThenElse",
        [([w_true, x, y], x),
         ([w_false, x, y], y)])

# fromTo m n | m > n = []
#            | otherwise = m : fromTo (m + 1) n
m = Var("m")
n = Var("n")

fromto = ForwardReference()
fromto.become(function("fromTo", [
    ([m, n], app(ifthenelse, app(prim_gt, m, n), w_nil,
                 cons(m, app(fromto, app(prim_add, m, Int(1)), n))))]))

from_ = ForwardReference()
from_.become(function("from",
        [([m], cons(m, app(from_, app(succ, m))))], recursive=True))

take = ForwardReference()
take.become(function("take",
        [([Int(0), xs], w_nil),
            ([n, w_nil], w_nil),
            ([n, cons(x, xs)], cons(x, app(take,
                app(prim_add, n, Int(-1)), xs)))], recursive=True))

# head (x : xs) = x

head = function("head", [
    ([cons(hd, tl)], hd)])

# last (x:xs) = last2 xs x
#
# last2 [] x = x
# last2 ys _ = last ys

last = ForwardReference()
last2 = ForwardReference()
last.become(function("last", [
    ([cons(hd, tl)], app(last2, tl, hd))]))
last2.become(function("last2", [
    ([w_nil, n], n),
    ([xs, m], app(last, xs))], recursive=True))


eq_int_list = ForwardReference()
eq_int_list.become(function("eq_int_list", [
    ([w_nil, w_nil], w_true),
    ([cons(x, xs), cons(y, ys)],
     app(and_, app(prim_eq, x, y), app(eq_int_list, xs, ys))),
    ([xs, ys], w_false)]))

# iterate f x = x : iterate f (f x)

iterate = ForwardReference()
iterate.become(function("iterate", [
    ([f, x], cons(x, app(iterate, f, app(f, x))))]))

# filter f []     = []
# filter f (x:xs) = if f x then x : filter f xs else filter f xs

filter_ = ForwardReference()
filter_.become(function("filter", [
    ([f, w_nil], w_nil),
    ([f, cons(x, xs)], app(ifthenelse, app(f, x),
                           cons(x, app(filter_, f, xs)),
                           app(filter_, f, xs)))],
    recursive=True))


def to_list(l):
    current = w_nil
    for element in reversed(l):
        if isinstance(element, str):
            element = constr(element)
        if isinstance(element, int):
            element = Int(element)
        current = cons(element, current)
    return current


def test_substitute():
    x = NumberedVar(0)
    subst = [constr("True")]
    assert x.substitute(subst) == constr("True")
    assert constr("Just", x).substitute(subst) == \
            constr("Just", constr("True"))

    appl = make_application(id_, [x])
    new_app = appl.substitute(subst)
    assert new_app.getarg(0) == constr("True")

    appl = make_application(x, [w_true])
    new_app = appl.substitute([id_])
    assert new_app.function is id_


def test_substitute_thunks():
    appl = app(id_, w_true)
    subst, todo = dup.apply(app(dup, appl), None)
    assert todo is None
    res = subst.subst[0]
    assert isinstance(res, Thunk)
    assert res.application == appl

    thunk1 = thunk(id_, w_true)
    subst, todo = dup.apply(app(dup, thunk1), None)
    assert todo is None
    res = subst.subst[0]
    assert res is thunk1

    thunk2 = thunk(None)
    thunk2.application = Int(42)
    subst, todo = dup.apply(app(dup, thunk2), None)
    assert todo is None
    res = subst.subst[0]
    assert res == Int(42)

    subst, todo = id_.apply(app(id_, appl), None)
    assert subst.subst[0] is appl

    subst, todo = id_.apply(app(id_, thunk1), None)
    assert subst.subst[0] is thunk1

    subst, todo = id_.apply(app(id_, thunk2), None)
    assert subst.subst[0] == Int(42)


def test_pattern_matching():
    x = NumberedVar(0)
    subst = [None]
    assert x.match(constr("True"), subst)
    assert subst == [constr("True")]

    p1 = constr("True")
    subst = []
    assert p1.match(p1, subst)
    assert subst == []
    assert not p1.match(constr("False"), subst)

    subst = [None]
    p = constr("Just", x)
    assert p.match(constr("Just", constr("True")), subst)
    assert subst == [constr("True")]
    assert not p.match(constr("False"), subst)

    subst = []
    assert p1.match(thunk(id_, constr("True")), subst) == NEEDS_HNF

    x = Var("x")
    g = function("g", [([x], app(id_, x))])
    assert p1.match(thunk(g, constr("True")), subst) == NEEDS_HNF
    assert subst == []


def test_occurrences():
    assert id_.rules[0].occurrences == [1]

    assert dup.rules[0].occurrences == [2]

    assert ifthenelse.rules[0].occurrences == [1, 0]
    assert ifthenelse.rules[1].occurrences == [0, 1]


def test_function():

    assert id_.getvalue() == id_
    assert w_true.getvalue() == w_true
    x = evaluate_hnf(make_application(id_, [w_true]))
    assert x == w_true

    # not True = False
    # not False = True

    x = evaluate_hnf(make_application(not_, [w_true]))
    assert x == w_false
    x = evaluate_hnf(make_application(not_, [w_false]))
    assert x == w_true


def test_sharing():
    result = evaluate_hnf(app(dup, constr("abc")))
    assert result.getarg(0) is result.getarg(1)

    th = thunk(id_, constr("True"))
    result = evaluate_hnf(app(dup, th))
    assert result.getarg(0) is result.getarg(1)


def test_sharing2():
    l = [0]

    @expose_primitive(1)
    def counting_id(args):
        l[0] += 1
        return args[0]

    x = evaluate_hnf(app(eq_int_list,
                         app(map_, app(counting_id, succ),
                             to_list([1, 2, 3])), to_list([2, 3, 4])))
    assert x == w_true
    assert l == [1]


def test_higher_order():
    x = evaluate_hnf(app(thunk(id_, not_), w_true))
    assert x == w_false


def test_and():
    x = evaluate_hnf(app(and_,
        w_true, thunk(id_, w_false)))
    assert x == w_false


def test_thunkthunk():
    f = function("myid", [
        ([x], app(id_, x))])
    res = evaluate_hnf(app(f, app(id_, w_true)))
    assert res == w_true


def test_functions_as_results():
    res = evaluate_hnf(app(app(ifthenelse, w_true, succ, id_), Int(42)))
    assert res == Int(43)


def test_eq_int_list():
    expr = app(eq_int_list,
        cons(Int(1), cons(Int(2), w_nil)),
        app(id_, cons(Int(1), cons(Int(2), w_nil))))
    x = evaluate_hnf(expr)
    assert x == w_true
    x = evaluate_hnf(app(eq_int_list,
        cons(Int(1), cons(Int(2), w_nil)),
        app(id_, cons(Int(1), cons(Int(3), w_nil)))))
    assert x == w_false
    x = evaluate_hnf(app(eq_int_list,
        cons(Int(1), cons(Int(2), w_nil)),
        app(id_, cons(Int(1), w_nil))))
    assert x == w_false


def test_append():
    x = evaluate_hnf(app(eq_int_list, to_list([1, 2, 3, 4, 5, 6]),
                     app(append, to_list([1, 2, 3]), to_list([4, 5, 6]))))
    assert x == w_true
    x = evaluate_hnf(app(eq_int_list, to_list([1] * 1000),
                     app(append, to_list([1] * 990), to_list([1] * 10))))
    assert x == w_true


def test_map():
    x = evaluate_hnf(app(eq_int_list, app(map_, succ, to_list([1, 2, 3])),
                                      to_list([2, 3, 4])))
    assert x == w_true


def test_tree_map():
    tree = node(node(w_empty, Int(1), w_empty), Int(2), w_empty)
    succtree = node(node(w_empty, Int(2), w_empty), Int(3), w_empty)
    value = evaluate_hnf(app(eq_int_tree, app(treemap, succ, tree), succtree))
    assert value == w_true


def test_finite_inc():
    """Test for pattern matching on integers."""
    finite_inc = function("finite-inc", [
        ([Int(0)], Int(1)),
        ([Int(1)], Int(2)),
        ([Int(2)], Int(3))])
    result = evaluate_hnf(make_application(finite_inc, [Int(1)]))
    assert result == Int(2)


def test_int_id():
    """Test for substitution of variables by an integer."""
    result = evaluate_hnf(make_application(id_, [Int(1)]))
    assert result == Int(1)


def test_add1():
    """Test for application of primitive addition."""
    result = evaluate_hnf(make_application(prim_add, [Int(1), Int(2)]))
    assert result == Int(3)


def test_add2():
    """Test for evaluation of the arguments of primitive addition."""
    result = evaluate_hnf(make_application(prim_add,
                           [app(succ, Int(1)), app(succ, Int(2))]))
    assert result == Int(5)


def test_length():
    """Test that calculates the length of a list."""
    test_list = to_list(["a", "a", "a"])
    result = evaluate_hnf(make_application(length_, [test_list]))
    assert result == Int(3)


def test_length_iter():
    """Test that calculates the length of a list iteratively."""
    test_list = to_list(["a", "a", "a"])
    result = evaluate_hnf(make_application(length_iter, [test_list, Int(0)]))
    assert result == Int(3)


def test_length_seq_iter():
    """"Test that calculates the length of a list.

    With strict iterative implementation.
    """
    # for a list with 200 elements the maximum recursion depth is reached
    test_list = to_list("a" * 100)
    result = evaluate_hnf(make_application(length_seq_iter,
                                           [test_list, Int(0)]))
    assert result == Int(100)


def test_ifthenelse():
    res = evaluate_hnf(make_application(ifthenelse, [w_true, Int(0), Int(1)]))
    assert res == Int(0)
    res = evaluate_hnf(make_application(ifthenelse, [w_false, Int(0), Int(1)]))
    assert res == Int(1)


def test_fromto():
    result = evaluate_hnf(app(eq_int_list, app(fromto, Int(10), Int(14)),
                              to_list([10, 11, 12, 13, 14])))
    assert result == w_true


def test_take_from():
    result = evaluate_hnf(
            app(eq_int_list, app(take, Int(10), app(from_, Int(1))),
                              app(fromto, Int(1), Int(10))))
    assert result == w_true


def test_head_reverse():
    for i in range(1, 100, 10):
        result = evaluate_hnf(app(head, app(reverse,
                                            app(fromto, Int(1), Int(i)))))
        assert result == Int(i)


def test_last_fromto():
    for i in range(1, 300, 100):
        result = evaluate_hnf(app(last, app(fromto, Int(1), Int(i))))
        assert result == Int(i)


def test_seq():
    x = app(prim_seq, Int(1), None)
    next, todo = x.step(None)
    assert next is None
    assert todo is None


def test_constructor_n():
    x = constr("Twelve", Int(1), Int(2), Int(3), Int(4), Int(5), Int(6),
               Int(7), Int(8), Int(9), Int(10), Int(11), Int(12))
    assert x.numargs() == 12
    assert x.getarg(11) == Int(12)
    assert x.getname() == "Twelve"


def test_application_n():
    x = make_application(not_, [Int(1), Int(2), Int(3), Int(4),
                                Int(5), Int(6), Int(7), Int(8),
                                Int(9), Int(10), Int(11), Int(12)])
    assert x.numargs() == 12
    assert x.getarg(11) == Int(12)
    assert str(x) == "(not ((Constr True)) ((Constr False))) "\
            "1 2 3 4 5 6 7 8 9 10 11 12"


def test_tostr():
    assert str(succ) == "(succ (x))"
    assert str(map_) == "(map (f (Constr Nil)) (f (Constr Cons hd, tl)))"
    assert str(value) == "Var: value"
    assert str(prim_add) == "f"
    assert str(map_.rules[0]) == "(f (Constr Nil))"

    a = constr("tostr", Int(1), Float(2.0), Double(3.0))
    assert str(a) == "(Constr tostr 1, 2.0, 3.0)"

    b = thunk(id_, not_)
    assert str(b) == "?(id (x)) (not ((Constr True)) ((Constr False)))"

    c = PartialApp(not_, [])
    assert str(c) == "(partialapp: (not ((Constr True)) ((Constr False))))"

    d = Function("Forward", None)
    assert str(d) == "ForwardReference(Forward)"
