import py
import pytest
import pyhaskell.main as pymain


IGNORE = ('quicksort', 'enumerations', 'letrec', 'guards',
          'listcomp', 'classes', 'case2')


def pytest_collect_file(path, parent):
    if path.ext in ('.hs', '.lhs'):
        return HaskellFile(path, parent)


class HaskellTestError(Exception):
    """Custom Haskell source code testing error."""
    pass


class HaskellFile(pytest.File):
    """Collect tests from a Haskell source file."""
    def collect(self):
        yield HaskellItem(self.fspath.purebasename, self)


class HaskellItem(pytest.Item):
    """A single test based on an Haskell source file."""

    def __init__(self, name, parent):
        super(HaskellItem, self).__init__(name, parent)
        self.path = parent.fspath

        # Skip tests which we do not support yet
        if self.path.purebasename in IGNORE:
            pytest.skip("test in list of ignored tests")

    def runtest(self):
        returncode, out, err = py.io.StdCaptureFD.call(
            pymain.main, ['test', self.path.strpath])

        assert returncode == 0

        # Check expect stdout or stderr output
        out_path = self.path.new(ext='.stdout')
        if out_path.check(file=True):
            assert self._test_output(out, out_path.readlines(cr=False))
        err_path = self.path.new(ext='.stderr')
        if err_path.check(file=True):
            assert self._test_output(err, err_path.readlines(cr=False))

    def _test_output(self, captured, expected):
        """Check if 'captured' output matches 'expected' output."""
        captured = captured.split('\n') if captured is not None else ['']

        if len(captured) > 2 and captured[0].startswith("[1 of 1] Compiling"):
            captured = captured[2:]

        if len(expected) != len(captured):
            raise HaskellTestError("Error: incorrect output")
        for i, line in enumerate(expected):
            if captured[i] != line:
                raise HaskellTestError("Error: Expected '%s' got '%s'" %
                                       (line, captured[i]))

        return True
